@@ -9,7 +9,14 @@
  * license.
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+
 #include <locale.h>
 #include <gtk/gtk.h>
 #include <glib/gi18n.h>
@@ -17,7 +24,13 @@
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <gdk/gdkx.h>
 
-#include "lightdm.h"
+#ifdef HAVE_LIBINDICATOR
+#include <libindicator/indicator-object.h>
+#endif
+
+#include <lightdm.h>
+
+#include <src/lightdm-gtk-greeter-ui.h>
 
 static LightDMGreeter *greeter;
 static GKeyFile *state;
@@ -26,15 +39,207 @@ static GtkWindow *login_window, *panel_window;
 static GtkButton *login_button, *cancel_button;
 static GtkLabel *message_label, *prompt_label;
 static GtkWidget *login_box, *prompt_box;
+static GtkImage *user_avatar;
 static GtkEntry *prompt_entry;
 static GtkComboBox *user_combo;
 static GtkComboBox *session_combo;
 static GtkComboBox *language_combo;
 static gchar *default_font_name, *default_theme_name, *default_icon_theme_name;
 static GdkPixbuf *default_background_pixbuf = NULL;
+#if GTK_CHECK_VERSION (3, 0, 0)
 static GdkRGBA *default_background_color = NULL;
+#else
+static GdkColor *default_background_color = NULL;
+#endif
 static gboolean cancelling = FALSE, prompted = FALSE;
 
+
+#ifdef HAVE_LIBINDICATOR
+static gboolean
+entry_scrolled (GtkWidget *menuitem, GdkEventScroll *event, gpointer data)
+{
+    IndicatorObject      *io;
+    IndicatorObjectEntry *entry;
+
+    g_return_val_if_fail (GTK_IS_WIDGET (menuitem), FALSE);
+
+    io = g_object_get_data (G_OBJECT (menuitem), "indicator-custom-object-data");
+    entry = g_object_get_data (G_OBJECT (menuitem), "indicator-custom-entry-data");
+
+    g_return_val_if_fail (INDICATOR_IS_OBJECT (io), FALSE);
+
+    g_signal_emit_by_name (io, "scroll", 1, event->direction);
+    g_signal_emit_by_name (io, "scroll-entry", entry, 1, event->direction);
+
+    return FALSE;
+}
+
+static void
+entry_activated (GtkWidget *widget, gpointer user_data)
+{
+    IndicatorObject      *io;
+    IndicatorObjectEntry *entry;
+
+    g_return_if_fail (GTK_IS_WIDGET (widget));
+
+    io = g_object_get_data (G_OBJECT (widget), "indicator-custom-object-data");
+    entry = g_object_get_data (G_OBJECT (widget), "indicator-custom-entry-data");
+
+    g_return_if_fail (INDICATOR_IS_OBJECT (io));
+
+    return indicator_object_entry_activate (io, entry, gtk_get_current_event_time ());
+}
+
+static GtkWidget*
+create_menuitem (IndicatorObject *io, IndicatorObjectEntry *entry, GtkWidget *menubar)
+{
+    GtkWidget *box, *menuitem;
+
+#if GTK_CHECK_VERSION (3, 0, 0)
+    box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 3);
+#else
+    box = gtk_hbox_new (FALSE, 0);
+#endif
+    menuitem = gtk_menu_item_new ();
+
+    gtk_widget_add_events(GTK_WIDGET(menuitem), GDK_SCROLL_MASK);
+
+    g_object_set_data (G_OBJECT (menuitem), "indicator-custom-box-data", box);
+    g_object_set_data (G_OBJECT (menuitem), "indicator-custom-object-data", io);
+    g_object_set_data (G_OBJECT (menuitem), "indicator-custom-entry-data", entry);
+
+    g_signal_connect (G_OBJECT (menuitem), "activate", G_CALLBACK (entry_activated), NULL);
+    g_signal_connect (G_OBJECT (menuitem), "scroll-event", G_CALLBACK (entry_scrolled), NULL);
+
+    if (entry->image)
+        gtk_box_pack_start (GTK_BOX(box), GTK_WIDGET(entry->image), FALSE, FALSE, 1);
+
+    if (entry->label)
+        gtk_box_pack_start(GTK_BOX(box), GTK_WIDGET(entry->label), FALSE, FALSE, 1);
+
+    if (entry->menu)
+        gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), GTK_WIDGET (entry->menu));
+
+    gtk_container_add (GTK_CONTAINER (menuitem), box);
+    gtk_widget_show (box);
+    gtk_menu_shell_append (GTK_MENU_SHELL (menubar), menuitem);
+
+    return menuitem;
+}
+
+static void
+entry_added (IndicatorObject *io, IndicatorObjectEntry *entry, gpointer user_data)
+{
+    GHashTable *menuitem_lookup;
+    GtkWidget  *menuitem;
+
+    /* if the menuitem doesn't already exist, create it now */
+    menuitem_lookup = g_object_get_data (G_OBJECT (io), "indicator-custom-menuitems-data");
+    g_return_if_fail (menuitem_lookup);
+    menuitem = g_hash_table_lookup (menuitem_lookup, entry);
+    if (!GTK_IS_WIDGET (menuitem))
+    {
+        menuitem = create_menuitem (io, entry, GTK_WIDGET (user_data));
+        g_hash_table_insert (menuitem_lookup, entry, menuitem);
+    }
+
+    gtk_widget_show (menuitem);
+}
+
+static void
+entry_removed_cb (GtkWidget *widget, gpointer userdata)
+{
+    IndicatorObject *io;
+    GHashTable      *menuitem_lookup;
+    GtkWidget       *menuitem;
+    gpointer         entry;
+
+    io = g_object_get_data (G_OBJECT (widget), "indicator-custom-object-data");
+    if (!INDICATOR_IS_OBJECT (io))
+        return;
+
+    entry = g_object_get_data (G_OBJECT (widget), "indicator-custom-entry-data");
+    if (entry != userdata)
+        return;
+
+    menuitem_lookup = g_object_get_data (G_OBJECT (io), "indicator-custom-menuitems-data");
+    g_return_if_fail (menuitem_lookup);
+    menuitem = g_hash_table_lookup (menuitem_lookup, entry);
+    if (GTK_IS_WIDGET (menuitem))
+        gtk_widget_hide (menuitem);
+
+    gtk_widget_destroy (widget);
+}
+
+static void
+entry_removed (IndicatorObject *io, IndicatorObjectEntry *entry, gpointer user_data)
+{
+    gtk_container_foreach (GTK_CONTAINER (user_data), entry_removed_cb, entry);
+}
+
+static void
+menu_show (IndicatorObject *io, IndicatorObjectEntry *entry, guint32 timestamp, gpointer user_data)
+{
+    IndicatorObjectEntry *entrydata;
+    GtkWidget            *menuitem;
+    GList                *entries, *lp;
+
+    menuitem = GTK_WIDGET (user_data);
+
+    if (!entry)
+    {
+        /* Close any open menus instead of opening one */
+        entries = indicator_object_get_entries (io);
+        for (lp = entries; lp; lp = g_list_next (entry))
+        {
+            entrydata = lp->data;
+            gtk_menu_popdown (entrydata->menu);
+        }
+        g_list_free (entries);
+
+        /* And tell the menuitem to exit activation mode too */
+        gtk_menu_shell_cancel (GTK_MENU_SHELL (menuitem));
+    }
+}
+
+static gboolean
+load_module (const gchar *name, GtkWidget *menubar)
+{
+    IndicatorObject *io;
+    GList           *entries, *lp;
+    gchar           *path;
+
+    g_return_val_if_fail (name, FALSE);
+
+    if (!g_str_has_suffix (name, G_MODULE_SUFFIX))
+        return FALSE;
+
+    path = g_build_filename (INDICATOR_DIR, name, NULL);
+    io = indicator_object_new_from_file (path);
+    g_free (path);
+
+    /* used to store/fetch menu entries */
+    g_object_set_data_full (G_OBJECT (io), "indicator-custom-menuitems-data",
+                            g_hash_table_new (g_direct_hash, g_direct_equal),
+                            (GDestroyNotify) g_hash_table_destroy);
+
+    g_signal_connect (G_OBJECT (io), INDICATOR_OBJECT_SIGNAL_ENTRY_ADDED,
+                      G_CALLBACK (entry_added), menubar);
+    g_signal_connect (G_OBJECT (io), INDICATOR_OBJECT_SIGNAL_ENTRY_REMOVED,
+                      G_CALLBACK (entry_removed), menubar);
+    g_signal_connect (G_OBJECT (io), INDICATOR_OBJECT_SIGNAL_MENU_SHOW,
+                      G_CALLBACK (menu_show), menubar);
+
+    entries = indicator_object_get_entries (io);
+    for (lp = entries; lp; lp = g_list_next (lp))
+        entry_added (io, lp->data, menubar);
+
+    g_list_free (entries);
+
+    return TRUE;
+}
+#endif
+
 static gchar *
 get_session ()
 {
@@ -194,6 +399,38 @@ set_user_background (const gchar *username)
         g_object_unref (bg);
 }
 
+static void
+set_user_image (const gchar *username)
+{
+    const gchar *path;
+    LightDMUser *user;
+    GdkPixbuf *image = NULL;
+    GError *error = NULL;
+
+    user = lightdm_user_list_get_user_by_name (lightdm_user_list_get_instance (), username);
+    if (user)
+    {
+        path = lightdm_user_get_image (user);
+        if (path)
+        {
+            image = gdk_pixbuf_new_from_file_at_scale (path, 64, 64, FALSE, &error);
+            if (image)
+            {
+                gtk_image_set_from_pixbuf (GTK_IMAGE (user_avatar), image);
+                g_object_unref (image);
+                return;
+            }
+            else
+            {
+                g_warning ("Failed to load user image: %s", error->message);
+                g_clear_error (&error);
+            }
+        }
+    }
+    /* otherwise, show the default avatar instead */
+    gtk_image_set_from_icon_name (GTK_IMAGE (user_avatar), "avatar-default", GTK_ICON_SIZE_DIALOG);
+}
+
 static void
 start_authentication (const gchar *username)
 {
@@ -345,6 +582,7 @@ user_combobox_active_changed_cb (GtkComboBox *widget, LightDMGreeter *greeter)
 
         set_login_button_label (user);
         set_user_background (user);
+        set_user_image (user);
         start_authentication (user);
         g_free (user);
     }
@@ -528,7 +766,7 @@ get_user_iter (const gchar *username, GtkTreeIter *iter)
     GtkTreeModel *model;
 
     model = gtk_combo_box_get_model (user_combo);
-  
+
     if (!gtk_tree_model_get_iter_first (model, iter))
         return FALSE;
     do
@@ -616,8 +854,8 @@ a11y_contrast_cb (GtkWidget *widget)
 {
     if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (widget)))
     {
-        g_object_set (gtk_settings_get_default (), "gtk-theme-name", "HighContrastInverse", NULL);
-        g_object_set (gtk_settings_get_default (), "gtk-icon-theme-name", "HighContrastInverse", NULL);
+        g_object_set (gtk_settings_get_default (), "gtk-theme-name", "HighContrast", NULL);
+        g_object_set (gtk_settings_get_default (), "gtk-icon-theme-name", "HighContrast", NULL);
     }
     else
     {
@@ -700,6 +938,7 @@ load_user_list ()
                 gtk_combo_box_set_active_iter (user_combo, &iter);
                 set_login_button_label (selected_user);
                 set_user_background (selected_user);
+                set_user_image (selected_user);
                 start_authentication (selected_user);
                 break;
             }
@@ -780,11 +1019,17 @@ set_background (GdkPixbuf *new_bg)
             {
                 GdkPixbuf *p = gdk_pixbuf_scale_simple (bg, monitor_geometry.width,
                                                         monitor_geometry.height, GDK_INTERP_BILINEAR);
+                if (!gdk_pixbuf_get_has_alpha (p))
+                    p = gdk_pixbuf_add_alpha (p, FALSE, 255, 255, 255);
                 gdk_cairo_set_source_pixbuf (c, p, monitor_geometry.x, monitor_geometry.y);
                 g_object_unref (p);
             }
             else
+#if GTK_CHECK_VERSION (3, 0, 0)
                 gdk_cairo_set_source_rgba (c, default_background_color);
+#else
+                gdk_cairo_set_source_color (c, default_background_color);
+#endif
             cairo_paint (c);
         }
 
@@ -808,8 +1053,18 @@ main (int argc, char **argv)
     GtkCellRenderer *renderer;
     GtkWidget *menuitem, *hbox, *image;
     gchar *value, *state_dir;
+#if GTK_CHECK_VERSION (3, 0, 0)
     GdkRGBA background_color;
+#else
+    GdkColor background_color;
+#endif
     GError *error = NULL;
+#ifdef HAVE_LIBINDICATOR
+    gchar **whitelist;
+    GDir *dir;
+    gsize length = 0;
+    guint indicators_loaded = 0, i;
+#endif
 
     /* Disable global menus */
     g_unsetenv ("UBUNTU_MENUPROXY");
@@ -856,7 +1111,11 @@ main (int argc, char **argv)
     value = g_key_file_get_value (config, "greeter", "background", NULL);
     if (!value)
         value = g_strdup ("#000000");
+#if GTK_CHECK_VERSION (3, 0, 0)
     if (!gdk_rgba_parse (&background_color, value))
+#else
+    if (!gdk_color_parse (value, &background_color))
+#endif
     {
         gchar *path;
         GError *error = NULL;
@@ -876,7 +1135,11 @@ main (int argc, char **argv)
     else
     {
         g_debug ("Using background color %s", value);
+#if GTK_CHECK_VERSION (3, 0, 0)
         default_background_color = gdk_rgba_copy (&background_color);
+#else
+        default_background_color = gdk_color_copy (&background_color);
+#endif
     }
     g_free (value);
 
@@ -928,7 +1191,8 @@ main (int argc, char **argv)
     g_free (value);
 
     builder = gtk_builder_new ();
-    if (!gtk_builder_add_from_file (builder, GREETER_DATA_DIR "/greeter.ui", &error))
+    if (!gtk_builder_add_from_string (builder, lightdm_gtk_greeter_ui,
+                                      lightdm_gtk_greeter_ui_length, &error))
     {
         g_warning ("Error loading UI: %s", error->message);
         return EXIT_FAILURE;
@@ -947,38 +1211,57 @@ main (int argc, char **argv)
     session_combo = GTK_COMBO_BOX (gtk_builder_get_object (builder, "session_combobox"));
     language_combo = GTK_COMBO_BOX (gtk_builder_get_object (builder, "language_combobox"));  
     panel_window = GTK_WINDOW (gtk_builder_get_object (builder, "panel_window"));
-
-    /* Load logo */
-    value = g_key_file_get_value (config, "greeter", "logo", NULL);
-    if (value)
-    {
-        gchar *path;
-        GtkImage *logo = GTK_IMAGE (gtk_builder_get_object (builder, "logo"));
-        GdkPixbuf *logo_pixbuf = NULL;
-        GError *error = NULL;
-
-        if (g_path_is_absolute (value))
-            path = g_strdup (value);
-        else
-            path = g_build_filename (GREETER_DATA_DIR, value, NULL);
-
-        g_debug ("Loading logo %s", path);
-        logo_pixbuf = gdk_pixbuf_new_from_file (path, &error);
-        if (logo_pixbuf)
-            gtk_image_set_from_pixbuf (logo, logo_pixbuf);
-        else
-           g_warning ("Failed to load logo: %s", error->message);
-        g_clear_error (&error);
-        g_object_unref (logo_pixbuf);
-        g_free (path);
-        g_free (value);
-    }
+    user_avatar = GTK_IMAGE (gtk_builder_get_object (builder, "user_avatar"));
 
     gtk_label_set_text (GTK_LABEL (gtk_builder_get_object (builder, "hostname_label")), lightdm_get_hostname ());
 
     /* Glade can't handle custom menuitems, so set them up manually */
+#ifdef HAVE_LIBINDICATOR
+    /* whitelisted indicator modules to show */
+    whitelist = g_key_file_get_string_list (config, "greeter", "show-indicators", &length, NULL);
+    menuitem = GTK_WIDGET (gtk_builder_get_object (builder, "menubar"));
+    /* load indicators */
+    if (g_file_test (INDICATOR_DIR, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)))
+    {
+        const gchar *name;
+        dir = g_dir_open (INDICATOR_DIR, 0, NULL);
+
+        while ((name = g_dir_read_name (dir)))
+        {
+            gboolean match = FALSE;
+            for (i = 0; i < length; ++i)
+                if ((match = (g_strcmp0 (name, whitelist[i]) == 0)))
+                    break;
+
+            if (G_LIKELY (!match))
+            {
+                g_debug ("Ignoring module (not in whitelist): %s", name);
+                continue;
+            }
+
+            if (load_module (name, menuitem))
+                ++indicators_loaded;
+        }
+
+        g_dir_close (dir);
+    }
+
+    if (length > 0)
+        g_strfreev (whitelist);
+
+    if (indicators_loaded > 0)
+    {
+        gtk_widget_set_can_focus (menuitem, TRUE);
+        gtk_widget_show (menuitem);
+    }
+#endif
+
     menuitem = GTK_WIDGET (gtk_builder_get_object (builder, "power_menuitem"));
+#if GTK_CHECK_VERSION (3, 0, 0)
     hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
+#else
+    hbox = gtk_hbox_new (FALSE, 0);
+#endif
     gtk_widget_show (hbox);
     gtk_container_add (GTK_CONTAINER (menuitem), hbox);
     image = gtk_image_new_from_icon_name ("system-shutdown", GTK_ICON_SIZE_MENU);
@@ -986,7 +1269,11 @@ main (int argc, char **argv)
     gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, TRUE, 0);
 
     menuitem = GTK_WIDGET (gtk_builder_get_object (builder, "a11y_menuitem"));
+#if GTK_CHECK_VERSION (3, 0, 0)
     hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
+#else
+    hbox = gtk_hbox_new (FALSE, 0);
+#endif
     gtk_widget_show (hbox);
     gtk_container_add (GTK_CONTAINER (menuitem), hbox);
     image = gtk_image_new_from_icon_name ("preferences-desktop-accessibility", GTK_ICON_SIZE_MENU);
@@ -1072,7 +1359,7 @@ main (int argc, char **argv)
         load_user_list ();
         gtk_widget_hide (GTK_WIDGET (cancel_button));
         gtk_widget_show (GTK_WIDGET (user_combo));
-    } 
+    }
 
     gtk_builder_connect_signals(builder, greeter);
 
@@ -1095,7 +1382,11 @@ main (int argc, char **argv)
     if (default_background_pixbuf)
         g_object_unref (default_background_pixbuf);
     if (default_background_color)
-        gdk_rgba_free ( default_background_color);
+#if GTK_CHECK_VERSION (3, 0, 0)
+        gdk_rgba_free (default_background_color);
+#else
+        gdk_color_free (default_background_color);
+#endif
 
     return EXIT_SUCCESS;
 }
